/*
        Copyright (c) 2012, MICEX-RTS. All rights reserved.

        Plaza-2 Client Gate API Usage Sample.
        Messages sender.

        All the software and documentation included in this and any
        other MICEX-RTS CGate Releasese is copyrighted by MICEX-RTS.

        Redistribution and use in source and binary forms, with or without
        modification, are permitted only by the terms of a valid
        software license agreement with MICEX-RTS.

        THIS SOFTWARE IS PROVIDED "AS IS" AND MICEX-RTS DISCLAIMS ALL WARRANTIES
        EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION, ANY IMPLIED WARRANTIES OF
        NON-INFRINGEMENT, MERCHANTABILITY OR FITNESS FOR A PARTICULAR
        PURPOSE.  MICEX-RTS DOES NOT WARRANT THAT USE OF THE SOFTWARE WILL BE
        UNINTERRUPTED OR ERROR-FREE.  MICEX-RTS SHALL NOT, UNDER ANY CIRCUMSTANCES, BE
        LIABLE TO LICENSEE FOR LOST PROFITS, CONSEQUENTIAL, INCIDENTAL, SPECIAL OR
        INDIRECT DAMAGES ARISING OUT OF OR RELATED TO THIS AGREEMENT OR THE
        TRANSACTIONS CONTEMPLATED HEREUNDER, EVEN IF MICEX-RTS HAS BEEN APPRISED OF
        THE LIKELIHOOD OF SUCH DAMAGES.
*/

/* 
	Send_mt -    . (:    ,  C+
	+11.).   1  .  2  reply   .

	Send_mt - a multistream example of sending order. (Attention! This example can be compiled only by compilers supporting C++11.)
	Stream 1 contains orders. Stream 2 contains replies for the orders sent.
*/

#include <thread>
#include <memory>
#include <vector>
#include <string>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <condition_variable>

#ifdef _WIN32
#include <windows.h>
#endif//_WIN32

//#include <assert.h>
#define CG_SHORT_NAMES

#include <cgate.h>
#include "cgate.hpp"
using namespace cgate; 
#include <cgutils.h>
#include <cgutils/rdtsc.h>

#include "send_mt_utils.h"
#include "router.h"
#include "router_path.h"

// helper variable used as a notification flag to exit
static volatile int bExit = 0;
int numLsns = 0;
static volatile int id = 0;

#ifdef _WIN32
BOOL WINAPI InterruptHandler(DWORD reason)
{
	printf("\n----BREAK----\n");
	bExit = 1;
	return 1;
}
#else
#include <unistd.h>
void Sleep(int msec)
{
	usleep(msec * 1000);
}
#endif//_WIN32

const uint32_t defaultTimeOut = 1;
const std::string lsn_str[] = 
{
	"p2ordbook://FORTS_ORDLOG_REPL;snapshot=FORTS_FUTORDERBOOK_REPL",
	"p2repl://FORTS_FUTINFO_REPL;scheme=|FILE|ini/fut_info.ini|CustReplScheme",
	"p2repl://FORTS_FUTTRADE_REPL",
	"p2repl://FORTS_FUTTRADE_REPL;scheme=|FILE|ini/fut_trades.ini|CustReplScheme",
	"p2ordbook://FORTS_FUTTRADE_REPL;snapshot=FORTS_FUTORDERBOOK_REPL;scheme=|FILE|ini/fut_trades.ini|CustReplScheme",
	"p2ordbook://FORTS_ORDLOG_REPL;snapshot=FORTS_FUTORDERBOOK_REPL;scheme=|FILE|ini/ordLog_trades.ini|CustReplScheme",
	"p2ordbook://FORTS_FUTTRADE_REPL;snapshot=FORTS_FUTORDERBOOK_REPL" 
};
const size_t max_lsns_cnt = 7;



template <int _port> class BlinkingRouterWrapper
{
	static Router _router;
	static const uint64_t WorkPeriod = 15 * 1000000; // usec
	static const uint64_t StandbyPeriod = 5 * 1000000; // usec

	static CG_RESULT startRouter()
	{
		TStrMap valueMap= create_map<std::string, std::string>("{routedefault}", "")("{routertype}", "root");
		int routerPort = -1;
		_router.start(routerPath, valueMap, routerPort);
		if (routerPort == 0)
			return CG_ERR_INTERNAL;
		return _router.wait(10);
	}

public:

	static CG_RESULT createAndRun()
	{
		CG_RESULT res = startRouter();
		if (res != CG_ERR_OK)
			return res;

		uint64_t turnStartTime = cg_usec();
		enum STATE { WORKING, STANDBY } state(WORKING);
		while (!bExit)
		{
			uint64_t currentTime = cg_usec();
			if ((state == WORKING) && (currentTime - turnStartTime >= WorkPeriod))
			{
				_router.stop();
				state = STANDBY;
			}
			else if ((state == STANDBY) && (currentTime - turnStartTime >= WorkPeriod + StandbyPeriod))
			{
				res = startRouter();
				if (res != CG_ERR_OK)
					return res;
				state = WORKING; 
				turnStartTime = cg_usec();
			}
		}
	}
};

Router BlinkingRouterWrapper<4001>::_router(4001);

class MessageListener : public Listener::IEventsListener
{
	std::unique_ptr<Connection> _conn;
	std::unique_ptr<Listener> _lsn;
	std::string _connSettings;
	std::string _lsnSettings;

public:
	MessageListener(const std::string& connSettings, const std::string& lsnSettings) 
		: _conn(new Connection(connSettings))
		, _lsn(new Listener(*_conn.get(), lsnSettings, this))
		, _connSettings(connSettings)
		, _lsnSettings(lsnSettings)
	{
	}

	inline void process()
	{
		State state = _conn->getState();
		if (state == State::Error)
		{
			_conn->close();
			_lsn.release();
		}
		if (state == State::Closed)
		{
			_conn->open("");
			_lsn.reset(new Listener(*_conn.get(), _lsnSettings, this));
		}
		if (state == State::Active)
		{
			_conn->process(defaultTimeOut);

			state = _lsn->getState();
			if (state == State::Error)
			{
				_lsn->close();
				_lsn.release();
				_conn->close();
				_conn.release();
				_conn.reset(new Connection(_connSettings));
				_lsn.reset(new Listener(*_conn.get(), _lsnSettings, this));
			}
			if (state == State::Closed)
				_lsn->open("");
		}
	}

	void run()
	{
		while(!bExit)
		{
			try
			{
				process();
			}
			catch(Exception& e)
			{
				/*if (e.code() != CG_ERR_TIMEOUT)
					CGate::logError("Listener run: cgate exception %d, %s", e.code(), e.what().c_str());*/
			}
		}
		CGate::logInfo("Listener: finished");
	}

	static bool pred()
	{
		return id == numLsns;
	}

	static void createAndRun(std::unique_ptr<MessageListener>* listener, std::mutex* cv_m, std::condition_variable* cv)
	{
		{
			std::unique_lock<std::mutex> lk(*cv_m);
			id++;
			(*cv).wait(lk, pred);
			(*cv).notify_all();
		}
		listener->get()->run();
	}

	uint32_t event(cg_conn_t* conn, cg_listener_t* listener, struct cg_msg_t* msg)
	{
		if (msg->type == CG_MSG_DATA)
			cg_msg_data_t* msgData = reinterpret_cast<cg_msg_data_t*>(msg);

		return CG_ERR_OK;
	}

private:
	MessageListener& operator=(const MessageListener&);
	MessageListener(const MessageListener&);
};


int main(int argc, char* argv[])
{	
#ifdef _WIN32
	SetConsoleCtrlHandler(InterruptHandler, 1);
#endif//_WIN32
	try
	{
		CGate env("ini=repl_mt.ini;key=11111111");
	
		if (argc > 1)
			numLsns = atoi(argv[1]);
		if (numLsns == 0 || numLsns > max_lsns_cnt)
		{
			CGate::logError("Please set number of listeners. N <= %d.\r\n", max_lsns_cnt);
			return 1;
		}
		
		/*CG_RESULT (*func)(void) = BlinkingRouterWrapper<4001>::createAndRun;

		std::thread routerThread(func);*/
		
		std::condition_variable cv;
		std::mutex cv_m;
		
		std::vector< std::unique_ptr<MessageListener> > listeners;
		for(int i = 0; i < numLsns; ++i)
		{
			std::string conn_str = "p2tcp://127.0.0.1:4001;app_name=test_repl_mt_" + std::to_string(i);
			listeners.push_back(std::unique_ptr<MessageListener>(new MessageListener(conn_str, lsn_str[i])));
		}

		std::vector<std::thread> threads;
		threads.reserve(numLsns);
		int id = 0;

		for(int i = 0; i < numLsns; ++i)
		{
			threads.push_back(std::thread(MessageListener::createAndRun, &listeners[i], &cv_m, &cv));
		}

		for(int i = 0; i < numLsns; ++i)
		{
			threads[i].join();
		}
		//routerThread.join();
		CGate::logInfo("All threads joined");
	}
	catch(Exception& e)
	{
		CGate::logError("Main: cgate exception %d, %s", e.code(), e.what().c_str());
	}
	return 0;
}
